﻿using System;
using System.Collections;
using System.IO;
using System.Text;
using GHI.Premium.IO;
using Microsoft.SPOT;
using Math = System.Math;

namespace GcRebootTest
{
   public class Program
   {
      public static void Main()
      {
         var ps = new PersistentStorage("SD");
         ps.MountFileSystem();

         var loader1 = new Loader();
         loader1.OnLoaded(@"\SD\File1.txt");
         // this GC is ok
         Debug.GC(true);

         var loader2 = new Loader();
         loader2.OnLoaded(@"\SD\File2.txt");
         // this GC alwas freezes/reboots
         Debug.GC(true);

         // it does not matter in which order the files are loaded or if I load the same file twice

         var loader3 = new Loader();
         loader3.OnLoaded(@"\SD\File3.txt");
         Debug.GC(true);

         ps.UnmountFileSystem();

         ps.Dispose();
      }
   }

   public class Loader
   {
      public void OnLoaded(string paramFileName)
      {
         _FirstParam = null;
         KmaParam lastParam = null;
         int nextStart = 0;
         int lastTelNo = 1;
         
         using (var sr = new StreamReader(paramFileName))
         {
            var split = new string[8];
            int n = 0;
            int next;
            bool isNewLine = false;
            var charBuffer = new char[128];
            int bufferPos = 0;

            while ((next = sr.Read()) != -1)
            {
               if (next == 13 || next == 10)
               {
                  if (isNewLine)
                  {
                     isNewLine = false;
                     continue;
                  }
                  isNewLine = true;
               }
               else
               {
                  isNewLine = false;
               }
               if (next == 9 || isNewLine)
               {
                  if (n < split.Length)
                  {
                     split[n] = new string(charBuffer, 0, bufferPos);
                  }
                  bufferPos = 0;
                  ++n;
                  if (isNewLine)
                  {
                     if (n == 8 && (split[0].Length == 0 || split[0][0] != ';'))
                     {
                        var param = new KmaParam(this);
                        param.SetCondition(split[1]);
                        if (split[2].Length == 0)
                        {
                           param.TelegramNo = lastTelNo;
                        }
                        else
                        {
                           param.TelegramNo = Int32.Parse(split[2]);
                           if (param.TelegramNo != lastTelNo)
                           {
                              nextStart = 0;
                              lastTelNo = param.TelegramNo;
                           }
                        }

                        if (split[3].Length == 0)
                        {
                           param.Start = nextStart;
                        }
                        else
                        {
                           param.Start = Int32.Parse(split[3]);
                        }
                        param.Length = Int32.Parse(split[4]);
                        nextStart = param.Start + param.Length;

                        param.SetFormat(split[5]);
                        param.SourceName = split[6];
                        param.ValueItemName = split[7];

                        if (lastParam == null)
                        {
                           _FirstParam = param;
                        }
                        else
                        {
                           lastParam.Next = param;
                        }
                        lastParam = param;
                     }
                     n = 0;
                  }
               }
               else
               {
                  if (bufferPos >= 128)
                  {
                     throw new ApplicationException("char buffer too short");
                  }
                  charBuffer[bufferPos++] = (char)next;
               }
            }
         }
      }

      private KmaParam _FirstParam;
   }

   internal class KmaParam
   {
      internal KmaParam Next;

      private abstract class Formater
      {
         public abstract string FormatValue(object value);
      }

      private class CopyFormater : Formater
      {
         public override string FormatValue(object value)
         {
            return value.ToString();
         }
      }

      private class NumberFormater : Formater
      {
         private readonly int _IntDigits;
         private readonly int _FractionalDigits;

         public NumberFormater(string format)
         {
            var split = format.Split(new[] { ',' }, 3);
            if (split.Length < 2)
            {
               throw new ApplicationException("Invalid number format: " + format);
            }
            try
            {
               if (format[0] == 'N')
               {
                  // left adjusted
                  _IntDigits = 1;
               }
               else
               {
                  _IntDigits = Int32.Parse(split[1]);
               }
               if (split.Length >= 3)
               {
                  _FractionalDigits = Int32.Parse(split[2]);
               }
            }
            catch (Exception ex)
            {
               throw new ApplicationException("Invalid number format: " + format + ": " + ex.Message);
            }
         }

         public override string FormatValue(object value)
         {
            if (_FractionalDigits == 0)
            {
               int n;
               if (value is double)
               {
                  n = (int)Math.Round((double)value);
               }
               else if (value is int)
               {
                  n = (int)value;
               }
               else
               {
                  throw new InvalidOperationException("Can not convert value " + value + " to number");
               }
               if (_IntDigits <= 1)
               {
                  return n.ToString();
               }
               var sb = new StringBuilder(_IntDigits);
               sb.Append(n);
               if (sb.Length < _IntDigits)
               {
                  sb.Insert(0, " ", _IntDigits - sb.Length);
               }
               return sb.ToString();
            }
            else
            {
               double d;
               if (value is double)
               {
                  d = (double)value;
               }
               else if (value is int)
               {
                  d = (int)value;
               }
               else
               {
                  throw new InvalidOperationException("Can not convert value " + value + " to number");
               }
               if (_IntDigits <= 1)
               {
                  return d.ToString("F" + _FractionalDigits);
               }
               var l = _IntDigits + _FractionalDigits + 1;
               var sb = new StringBuilder(l);
               sb.Append(d.ToString("F" + _FractionalDigits));
               if (sb.Length < l)
               {
                  sb.Insert(0, " ", l - sb.Length);
               }
               return sb.ToString();
            }
         }
      }

      private class BoolFormater : Formater
      {
         private readonly string _TrueValue;
         private readonly string _FalseValue;

         public BoolFormater(string format)
         {
            var split = format.Split(new[] { ',' }, 3);
            if (split.Length < 2)
            {
               _TrueValue = "True";
            }
            else
            {
               _TrueValue = split[1];
            }
            if (split.Length < 3)
            {
               _FalseValue = "False";
            }
            else
            {
               _FalseValue = split[2];
            }
         }

         public override string FormatValue(object value)
         {
            if (!(value is bool))
            {
               throw new InvalidOperationException("Can not convert value " + value + " to bool");
            }
            return ((bool)value) ? _TrueValue : _FalseValue;
         }
      }

      /*private class DateTimeFormater : Formater
      {
         private readonly DateTimeParser _SourceParser;
         private readonly string _TargetFormat;

         public DateTimeFormater(string format)
         {
            var split = format.Split(new[] { ',' }, 3);
            if (split.Length < 3)
            {
               throw new ApplicationException("Invalid date time format: " + format);
            }
            _SourceParser = new DateTimeParser(split[1]);
            _TargetFormat = split[2];
         }

         public override string FormatValue(object value)
         {
            if (!(value is string))
            {
               throw new InvalidOperationException("Can not convert value " + value + " to date time");
            }
            return _SourceParser.ParseExact((string)value).ToString(_TargetFormat);
         }
      }*/

      public KmaParam(Loader node)
      {
         _Node = node;
      }

      static KmaParam()
      {
         _FormatCache.Add("", new CopyFormater());
      }

      private readonly Loader _Node;

      //public string IpmTagName;
      private bool _ConditionValue;
      //private ExpressionParser _ConditionExpression;

      public int TelegramNo;
      public int Start;
      public int Length;
      public string SourceName;
      public string ValueItemName;

      //private IValue _ValueItem;

      private Formater _Formater;

      public void SetCondition(string value)
      {
         if (value.ToLower().Equals("true"))
         {
            _ConditionValue = true;
         }
         else if (value.ToLower().Equals("false"))
         {
            _ConditionValue = false;
         }
         else
         {
            /*_ConditionExpression = new ExpressionParser(GetVarTypeCallback, GetVarValueCallback);
            _ConditionExpression.ParseExpression(value);*/
         }
      }

      /*private Type GetVarTypeCallback(string varName)
      {
         var valueField = _Node.FindNamedItem(varName) as IValue;
         if (valueField != null)
         {
            return valueField.GetValueType();
         }
         return null;
      }

      private object GetVarValueCallback(string varName, object context)
      {
         var ni = _Node.FindNamedItem(varName);
         short sessionId = (short)(context ?? -1);
         if (sessionId >= 0)
         {
            var sessionValue = ni as ISessionValue;
            if (sessionValue != null)
            {
               return sessionValue.EvaluateSessionValue(sessionId);
            }
         }
         var value = ni as IValue;
         if (value != null)
         {
            return value.EvaluateValue();
         }
         return null;
      }*/

      private static readonly Hashtable _FormatCache = new Hashtable();

      public void SetFormat(string format)
      {
         // ReSharper disable once ReplaceWithStringIsNullOrEmpty
         if (format != null)
         {
            _Formater = _FormatCache[format] as Formater;
            if (_Formater == null)
            {
               format = format.Trim();
               if (format.Length > 0)
               {
                  if (format[0] == 'N' || format[0] == 'n')
                  {
                     _Formater = new NumberFormater(format);
                  }
                  else if (format[0] == 'B' || format[0] == 'b')
                  {
                     _Formater = new BoolFormater(format);
                  }
                  /*else if (format.Length > 8 && format.Substring(0, 8).ToLower().Equals("datetime"))
                  {
                     _Formater = new DateTimeFormater(format);
                  }*/
               }
               if (_Formater != null)
               {
                  _FormatCache.Add(format, _Formater);
               }
            }
         }
         if (_Formater == null)
         {
            _Formater = _FormatCache[""] as Formater;
         }
      }

      /*public void InsertValue(short sessionId, byte[] tel)
      {
         if (_ValueItem == null)
         {
            return;
         }
         object value;
         var sessionValue = _ValueItem as ISessionValue;
         if (sessionValue != null)
         {
            value = sessionValue.EvaluateSessionValue(sessionId);
         }
         else
         {
            value = _ValueItem.EvaluateValue();
         }
         if (value != null)
         {
            try
            {
               var valueStr = _Formater.FormatValue(value);
               for (int n = 0; n < valueStr.Length && n < Length; ++n)
               {
                  tel[2 + Start + n] = (byte)valueStr[n];
               }
            }
            catch (Exception ex)
            {
               _Node.IoHandler.Log.LogException(ex, "Format value '", value, "' (T:", TelegramNo, " S:", Start, ")");
            }
         }
      }*/

      /*public void CheckCondition()
      {
         if (_ConditionExpression != null && !_ConditionExpression.CheckReturnType(typeof(bool)))
         {
            throw new ApplicationException("The return type of the condition is not bool");
         }
      }

      public bool EvaluateCondition(short sessionId)
      {
         if (_ConditionExpression != null)
         {
            return _ConditionExpression.EvaluateBool(sessionId);
         }
         return _ConditionValue;
      }

      public void ResolveValueItem(ref INamedItem source)
      {
         if (ValueItemName == null || ValueItemName.Length == 0)
         {
            return;
         }
         if (source == null && (SourceName == null || SourceName.Length == 0))
         {
            throw new ApplicationException("Source is not defined");
         }
         if (SourceName != null && SourceName.Length > 0)
         {
            // ReSharper disable once SuspiciousTypeConversion.Global
            if (source == null || !SourceName.Equals((string)source.FullName))
            {
               source = _Node.FindNamedItem(SourceName);
               if (source == null)
               {
                  throw new ApplicationException("Could not find source item " + SourceName);
               }
            }
            SourceName = null; // release some mem
         }
         var ds = source as DataSource;
         if (ds != null)
         {
            _ValueItem = ds.FindField(ValueItemName);
         }
         else
         {
            _ValueItem = source.FindNamedItem(ValueItemName) as IValue;
         }
         if (_ValueItem == null)
         {
            throw new ApplicationException("Could not find item " + ValueItemName + " or item is not a value");
         }
         ValueItemName = null; // release some mem
      }*/
   }
}
